@@ -44,6 +44,108 @@
 #include "magick/tempfile.h"
 #include "magick/utility.h"
 
+/*
+  Forward declarations.
+*/
+static void DestroyImageAttribute(ImageAttribute *attribute);
+
+/*
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+%                                                                             %
+%                                                                             %
+%                                                                             %
+%   C l o n e I m a g e A t t r i b u t e s                                   %
+%                                                                             %
+%                                                                             %
+%                                                                             %
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+%
+%  CloneImageAttributes() copies the text attibutes from one image to another.
+%  Any text attributes in the destination image are preserved.
+%  CloneImageAttributes returns MagickPass if all of the attribututes are
+%  successfully cloned or MagickFail if there is a memory allocation error.
+%
+%  The format of the CloneImageAttributes method is:
+%
+%      MagickPassFail CloneImageAttributes(Image* clone_image,
+%                                          const Image* original_image)
+%
+%  A description of each parameter follows:
+%
+%    o clone_image: The destination image.
+%
+%    o original_image: The source image.
+%
+%
+*/
+MagickExport MagickPassFail CloneImageAttributes(Image* clone_image,
+  const Image* original_image)
+{
+  MagickPassFail
+    status;
+
+  ImageAttribute
+    *cloned_attribute,
+    *cloned_attributes;
+
+  const ImageAttribute
+    *attribute;
+
+  status = MagickPass;
+
+  /*
+    Search for tail of list (if any)
+  */
+  for(cloned_attributes=clone_image->attributes;
+      cloned_attributes != (ImageAttribute *) NULL;
+      cloned_attributes=cloned_attributes->next);
+
+  attribute=GetImageAttribute(original_image,(char *) NULL);
+  for ( ; attribute != (const ImageAttribute *) NULL; attribute=attribute->next)
+    {
+      /*
+        Construct AttributeInfo to append.
+      */
+      cloned_attribute=MagickAllocateMemory(ImageAttribute *,sizeof(ImageAttribute));
+      if (cloned_attribute == (ImageAttribute *) NULL)
+        {
+          status = MagickFail;
+          break;
+        }
+      cloned_attribute->key=AcquireString(attribute->key);
+      cloned_attribute->value=AcquireString(attribute->value);
+      cloned_attribute->previous=(ImageAttribute *) NULL;
+      cloned_attribute->next=(ImageAttribute *) NULL;
+      if ((cloned_attribute->value == (char *) NULL) ||
+          (cloned_attribute->key == (char *) NULL))
+        {
+          DestroyImageAttribute(cloned_attribute);
+          status = MagickFail;
+          break;
+        }
+
+      if (cloned_attributes == (ImageAttribute *) NULL)
+        {
+          /*
+            Start list
+          */
+          cloned_attributes=cloned_attribute;
+          clone_image->attributes=cloned_attributes;
+        }
+      else
+        {
+          /*
+            Append to list
+          */
+          cloned_attributes->next=cloned_attribute;
+          cloned_attribute->previous=cloned_attributes;
+          cloned_attributes=cloned_attribute;
+        }
+    }
+
+  return status;
+}
+
 /*
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %                                                                             %
@@ -68,6 +170,15 @@
 %
 %
 */
+static void DestroyImageAttribute(ImageAttribute *attribute)
+{
+  if (attribute == (ImageAttribute *) NULL)
+    return;
+  MagickFreeMemory(attribute->value);
+  MagickFreeMemory(attribute->key);
+  (void) memset(attribute,0xbf,sizeof(ImageAttribute));
+  MagickFreeMemory(attribute);
+}
 MagickExport void DestroyImageAttributes(Image *image)
 {
   ImageAttribute
@@ -82,11 +193,7 @@ MagickExport void DestroyImageAttributes(Image *image)
   {
     attribute=p;
     p=p->next;
-    if (attribute->key != (char *) NULL)
-      MagickFreeMemory(attribute->key);
-    if (attribute->value != (char *) NULL)
-      MagickFreeMemory(attribute->value);
-    MagickFreeMemory(attribute);
+    DestroyImageAttribute(attribute);
   }
   image->attributes=(ImageAttribute *) NULL;
 }
@@ -989,13 +1096,13 @@ static int GenerateEXIFAttribute(Image *image,const char *specification)
       }
   }
   if (index < 0)
-    return(False);
+    goto generate_attribute_failure;
   /*
     If EXIF data exists, then try to parse the request for a tag.
   */
   key=(char *) &specification[5];
   if ((key == (char *) NULL) || (*key == '\0'))
-    return(False);
+    goto generate_attribute_failure;
   while (isspace((int) (*key)))
     key++;
   all=0;
@@ -1032,7 +1139,7 @@ static int GenerateEXIFAttribute(Image *image,const char *specification)
       key++;
       n=strlen(key);
       if (n != 4)
-        return(False);
+        goto generate_attribute_failure;
       else
         {
           /*
@@ -1054,7 +1161,7 @@ static int GenerateEXIFAttribute(Image *image,const char *specification)
                   if ((c >= 'a') && (c <= 'f'))
                     tag|=c-('a'-10);
                   else
-                    return(False);
+                    goto generate_attribute_failure;
             }
           } while (*key != '\0');
         }
@@ -1079,7 +1186,7 @@ static int GenerateEXIFAttribute(Image *image,const char *specification)
     }
   }
   if (tag < 0)
-    return(False);
+    goto generate_attribute_failure;
   length=image->generic_profile[index].length;
   info=image->generic_profile[index].info;
   while (length != 0)
@@ -1099,7 +1206,7 @@ static int GenerateEXIFAttribute(Image *image,const char *specification)
     break;
   }
   if (length < 16)
-    return(False);
+    goto generate_attribute_failure;
   tiffp=info;
   id=Read16u(0,tiffp);
   morder=0;
@@ -1109,15 +1216,15 @@ static int GenerateEXIFAttribute(Image *image,const char *specification)
     if (id == 0x4D4D) /* MSB */
       morder=1;
     else
-      return(False);
+      goto generate_attribute_failure;
   if (Read16u(morder,tiffp+2) != 0x002a)
-    return(False);
+    goto generate_attribute_failure;
   /*
     This is the offset to the first IFD.
   */
   offset=Read32u(morder,tiffp+4);
   if (offset >= length)
-    return(False);
+    goto generate_attribute_failure;
   /*
     Set the pointer to the first IFD and follow it were it leads.
   */
@@ -1307,11 +1414,11 @@ static int GenerateEXIFAttribute(Image *image,const char *specification)
         }
         if ((t == TAG_EXIF_OFFSET) || (t == TAG_INTEROP_OFFSET))
           {
-            long
+            size_t
               offset;
 
-            offset=(long) Read32u(morder,pval);
-            if ((offset < (long) length) || (level < (DE_STACK_SIZE-2)))
+            offset=(size_t) Read32u(morder,pval);
+            if ((offset < length) && (level < (DE_STACK_SIZE-2)))
               {
                 /*
                   Push our current directory state onto the stack.
@@ -1336,6 +1443,9 @@ static int GenerateEXIFAttribute(Image *image,const char *specification)
   (void) SetImageAttribute(image,specification,(const char *) final);
   MagickFreeMemory(final);
   return(True);
+ generate_attribute_failure:
+  MagickFreeMemory(final);
+  return False;
 }
 
 MagickExport const ImageAttribute *GetImageAttribute(const Image *image,
