@@ -103,27 +103,39 @@
 %
 */
 
-static inline PixelPacket AlphaComposite(const PixelPacket *p,
-  const double alpha,const PixelPacket *q,const double beta)
+#define  MagickAlphaCompositeQuantum(change,change_alpha,base,base_alpha) \
+   (1.0-(change_alpha/MaxRGB))*(double) change+(1.0-(base_alpha/MaxRGB))*(double) base*(change_alpha/MaxRGB)
+static inline PixelPacket AlphaComposite(const PixelPacket *change,
+  const double change_alpha,const PixelPacket *base,const double base_alpha)
 {
   PixelPacket
     composite;
 
   double
-    MaxRGB_alpha,
-    MaxRGB_beta;
+    gamma,
+    value;
 
-  MaxRGB_alpha=MaxRGB-alpha;
-  MaxRGB_beta=MaxRGB-beta;
-  composite.red=(Quantum)
-    ((MaxRGB_alpha*p->red+alpha*MaxRGB_beta*q->red/MaxRGB)/MaxRGB+0.5);
-  composite.green=(Quantum)
-    ((MaxRGB_alpha*p->green+alpha*MaxRGB_beta*q->green/MaxRGB)/MaxRGB+0.5);
-  composite.blue=(Quantum)
-    ((MaxRGB_alpha*p->blue+alpha*MaxRGB_beta*q->blue/MaxRGB)/MaxRGB+0.5);
-  composite.opacity=(Quantum)
-    (MaxRGB-(MaxRGB_alpha+alpha*MaxRGB_beta/MaxRGB)+0.5);
-  return(composite);
+  composite=*base;
+  if (change_alpha != TransparentOpacity)
+    {
+      gamma=1.0-(change_alpha/MaxRGB)*(base_alpha/MaxRGB);
+      
+      value=MaxRGB*(1.0-gamma);
+      composite.opacity=RoundSignedToQuantum(value);
+      
+      gamma=1.0/(gamma <= MagickEpsilon ? 1.0 : gamma);
+      
+      value=gamma*MagickAlphaCompositeQuantum(change->red,change_alpha,composite.red,base_alpha);
+      composite.red=RoundSignedToQuantum(value);
+      
+      value=gamma*MagickAlphaCompositeQuantum(change->green,change_alpha,composite.green,base_alpha);
+      composite.green=RoundSignedToQuantum(value);
+      
+      value=gamma*MagickAlphaCompositeQuantum(change->blue,change_alpha,composite.blue,base_alpha);
+      composite.blue=RoundSignedToQuantum(value);
+    }
+
+  return composite;
 }
 
 MagickExport unsigned int ColorFloodfillImage(Image *image,
@@ -177,12 +189,12 @@ MagickExport unsigned int ColorFloodfillImage(Image *image,
   */
   if (FuzzyColorMatch(&draw_info->fill,&target,image->fuzz))
     return(False);
-  floodplane=(unsigned char *) AcquireMemory(image->columns*image->rows);
-  segment_stack=(SegmentInfo *) AcquireMemory(MaxStacksize*sizeof(SegmentInfo));
+  floodplane=MagickAllocateMemory(unsigned char *,image->columns*image->rows);
+  segment_stack=MagickAllocateMemory(SegmentInfo *,MaxStacksize*sizeof(SegmentInfo));
   if ((floodplane== (unsigned char *) NULL) ||
       (segment_stack == (SegmentInfo *) NULL))
-    ThrowBinaryException(ResourceLimitError,"MemoryAllocationFailed",
-      "UnableToFloodfillImage");
+    ThrowBinaryException3(ResourceLimitError,MemoryAllocationFailed,
+      UnableToFloodfillImage);
   (void) memset(floodplane,False,image->columns*image->rows);
   /*
     Push initial segment on stack.
@@ -339,8 +351,8 @@ MagickExport unsigned int ColorFloodfillImage(Image *image,
           break;
       }
     }
-  LiberateMemory((void **) &segment_stack);
-  LiberateMemory((void **) &floodplane);
+  MagickFreeMemory(segment_stack);
+  MagickFreeMemory(floodplane);
   return(True);
 }
 
@@ -433,10 +445,10 @@ MagickExport unsigned int MatteFloodfillImage(Image *image,
   /*
     Allocate segment stack.
   */
-  segment_stack=(SegmentInfo *) AcquireMemory(MaxStacksize*sizeof(SegmentInfo));
+  segment_stack=MagickAllocateMemory(SegmentInfo *,MaxStacksize*sizeof(SegmentInfo));
   if (segment_stack == (SegmentInfo *) NULL)
-    ThrowBinaryException(ResourceLimitError,"MemoryAllocationFailed",
-      "UnableToFloodfillImage");
+    ThrowBinaryException3(ResourceLimitError,MemoryAllocationFailed,
+      UnableToFloodfillImage);
   /*
     Push initial segment on stack.
   */
@@ -536,7 +548,7 @@ MagickExport unsigned int MatteFloodfillImage(Image *image,
       start=x;
     } while (x <= x2);
   }
-  LiberateMemory((void **) &segment_stack);
+  MagickFreeMemory(segment_stack);
   return(True);
 }
 
